#!/usr/bin/perl
#
# XData.pm:
# Perl-language interface to `xdata'.
#
# Copyright (c) 2001 Chris Lightfoot <chris@ex-parrot.com>
# 
# This program is free software; you can redistribute it and/or modify it
# under the same terms as Perl itself.
#
# $Id: XData.pm,v 1.6 2001/09/15 15:13:30 chris Exp $
#


# POD begins....

=head1 NAME

Data::XData - perl interface to the `xdata' simple database system

=head1 SYNOPSIS

    use Data::XData;
    
    # Open the database for writing
    $xd = new Data::XData($path, $type, 1) or die "oops";
    
    # What keys exist?
    @keys = $xd->keys();
    foreach $k (@keys) {
        print "key $k\n";
        # Tell me about this key.
        my ($time, $size) = $xd->stat($k);
        print "  size $size, last modified "
              . scalar(localtime($time)) . "\n";
        # Obtain the data for this key.
        print "  data:\n";
        print $xd->get($k);
    }
    
    # Change existing or insert new key.
    $xd->put("fish", "Underwater swimming creatures", 1);
    
    # Delete key if it exists.
    if ($xd->exists("salmon")) {
        $xd->delete("salmon");
        print STDERR "deleted too-specific fishtype\n";
    }

=head1 DESCRIPTION

Object-oriented interface to xdata, allowing perl programs to use the
facilities provided by F<xdget>, F<xdput>, F<xdls> etc., without the overhead
of calling the individual programs.

Xdata is built on top of GDBM, and this library uses GDBM_File to do its dirty
work.

=head1 PUBLIC INTERFACE

=over 4

=cut

package Data::XData;
use GDBM_File;
use strict;
use vars qw($VERSION $Sizeof_time_t);

$VERSION = 0.20;

# FIXME assumes sizeof(int) == sizeof(time_t)
$Sizeof_time_t = length(pack('i', 42));

=item new PATH, TYPE, [WRITE, [CREATE]]

I<Class method.>
Create a new xdata object, associated with the given PATH (a colon-separated
list of directories) and type. If WRITE is true, the database will be opened
for writing. If CREATE is also true, then, if the database does not exist, it
will be created in the directory given by the first component of PATH.

=cut
sub new ($$$;$$) {
    my ($proto, $path, $type, $write, $create) = @_;
    my $class = ref($proto) || $proto;
    my @paths = split(':', $path);
    my %data;
    my $self = { };
    bless($self, $class);

    $write ||= 0;
    $create ||= 0;

    $self->{path} = $path;
    $self->{type} = $type;

    # Escape characters in the type.
    $type =~ s#=#==#g;
    $type =~ s#/#=_#g;

    # Attempt to find the data somewhere.
    foreach my $p (@paths) {
        tie(%data, 'GDBM_File', "$p/$type", ($write ? GDBM_WRITER : GDBM_READER), 0666);
        last if (tied %data);
    }

    # Create database, if applicable.
    if (!tied %data and $create) {
        tie(%data, 'GDBM_File', "$paths[0]/$type", GDBM_WRCREAT, 0666);
    }

    # OK, that worked.
    if (tied %data) {
        $self->{data} = \%data;
        return $self;
    }
    
    # Oops.
    return undef;
}

=item close

I<Instance method.>
Untie the object from its database. This is done automatically at object
destruction time, but you may wish to call it explicitly under some
circumstances.

=cut
sub close ($) {
    my $self = shift;
    untie(%{$self->{data}});
    $self->{data} = undef;
}

=item keys [INTERNAL]

I<Instance method.>
Returns, in list context, a list of the keys in the given database; if
INTERNAL is true, then internal keys (named C<$$:...>) are returned, instead
of `normal' keys.

=cut
sub keys ($;$) {
    my ($self, $internal) = @_;
    my @keys = keys(%{$self->{data}});
    if ($internal) {
        return map { s/\0$//; $_ } grep {  /^\$\$\:/ } @keys;
    } else {
        return map { s/\0$//; $_ } grep { !/^\$\$\:/ } @keys;
    }
    return map { substr($_, 0, length($_) - 1) } @keys;
}

=item stat KEY

I<Instance method.>
Returns, in list context, the modification time and length of the given KEY.
The modification time is returned in seconds since the epoch, and the length
in bytes. If the key is not found, it returns the empty list.

=cut
sub stat ($$) {
    my ($self, $key) = @_;
    my $datum = $self->{data}->{"$key\0"};
    if (defined($datum)) {
        my ($mtime, $length);
        # FIXME assumes sizeof(int) == sizeof(time_t)
        $length = length($datum) - $Sizeof_time_t;
        $mtime = unpack('i', $datum);
        return ($mtime, $length);
    } else {
        return ();
    }
}

=item get KEY

I<Instance method.>
Returns the data associated with the given KEY, or undef if the key does not
exist.

=cut
sub get ($$) {
    my ($self, $key) = @_;
    my $datum = $self->{data}->{"$key\0"};
    if (defined($datum)) {
        return substr($datum, $Sizeof_time_t);
    } else {
        return undef;
    }
}

=item put KEY, DATUM, [REPLACE]

I<Instance method.>
Inserts DATUM for KEY, replacing existing data if REPLACE is true. Returns
true if data were inserted, or undef otherwise.

=cut
sub put ($$$;$) {
    my ($self, $key, $datum, $replace) = @_;
    $replace ||= 0;
    if ($replace or !exists($self->{data}->{"$key\0"})) {
        my $newdatum = pack('i', time());
        $newdatum .= $datum;
        $self->{data}->{"$key\0"} = $newdatum;
        return 1;
    } else {
        return undef;
    }
}

=item exists KEY

I<Instance method.>
Returns true if KEY exists in the database.

=cut
sub exists($$) {
    my ($self, $key);
    return exists($self->{data}->{"$key\0"});
}

=item delete KEY

I<Instance method.>
Deletes KEY from the database, if it exists. If KEY is not an internal key
(C<$$:...>), then a deletion record C<$$:d:KEY> is created to record the
deletion.

=cut
sub delete($$) {
    my ($self, $key);
    return undef if (!exists($self->{data}->{"$key\0"}));
    $self->put("\$\$:d:$key", '', 1) if ($key !~ /^\$\$:/);
    delete $self->{data}->{"$key\0"};
    return 1;
}

=item parse_headers SCALAR

I<Class method.>
Parses SCALAR as a list of RFC822-style headers, returning a reference to a
hash of C<header name -- [values]>. Header names are folded to lower case.

This is not a core part of xdata, but is a convenience method to allow easy
use of an easy-to-parse, familiar and human-readable data format within data.

=cut
sub parse_headers ($) {
    my $hdr = shift;
    my ($res, $k, $v);
    my @lines = split("\n", $hdr);
    
    foreach my $l (@lines) {
        if ($l =~ /^([^\s:]+):\s*(.*)\s*$/) {
            push(@{$res->{lc($k)}}, $v) if (defined($k) and defined($v));
            $k = $1;
            $v = $2;
        } elsif ($l =~ /^\s+([^\s].*)\s*$/) {
            $v .= "\n$1";
        } elsif ($l =~ /^\s*$/) {
            last;
        }
    }
    
    push(@{$res->{lc($k)}}, $v) if (defined($k) and defined($v));

    return $res;
}

1;

__END__

=head1 COPYING

Copyright (c) 2001 Chris Lightfoot, <chris@ex-parrot.com>
F<http://www.ex-parrot.com/~chris/>

This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 BUGS

This module assumes that C<sizeof(time_t) = sizeof(int)>, in the C language
sense. This is true on most architectures (i386, UltraASPARC, AXP Alpha) but
it may not be true on yours.

=head1 VERSION

$Id: XData.pm,v 1.6 2001/09/15 15:13:30 chris Exp $

=head1 SEE ALSO

F<GDBM_File>, L<perlfunc/tie>; L<xdata(1)>, etc.

F<http://www.ex-parrot.com/~chris/xdata/>.

=cut

# alternative parse_headers implementation

*=item parse_headers STRING

I<Class method.>
Parses RFC822-ish headers. This is not a core part of xdata, but is a
convenience method to allow easy use of an easy-to-parse, familiar and
human-readable data format within data.

Returns, in list context, a reference to a hash of the headers and the
remainder of the text in STRING (separated from the headers by a blank line).
Header names are converted to lower case in the hash; if a header is repeated,
then the values of each line are concatenated with "\n", and folded header
lines are unfolded, replacing repeated whitespace at the fold by a single
space. So,

    Name: Fred
        Flinstone
    Email: <fred@bedrock.com>
    Phone: (345) 555 1234 (work)
    Phone: (791) 555 8965 (mobile)
    
    Fred Flintstone works in a quarry.

will yield

    ( { Name  => "Fred Flintstone",
        Email => "<fred@bedrock.com>",
        Phone => "(345) 555 1234 (work)\n(791) 555 8965 (mobile)" },
      "Fred Flintstone works in a quarry.\n")

=cut
sub parse_headers ($) {
    my $input = shift;
    my $hdr = { };
    my $txt = "";

    my @lines = split("\n", $input);

    my $h = 0;
    my ($h_name, $h_val);
    my $i;
    for ($i = 0; $i < scalar(@lines); ++$i) {
        last if ($lines[$i] =~ /^\s*$/); # end of headers
        if ($lines[$i] =~ /^([^ :]+):\s*(.*)/) {
            # new header.
            if (defined($h_name) and defined($h_val)) {
                $hdr->{lc($h_name)} .= "\n" if (exists($hdr->{lc($h_name)}));
                $hdr->{lc($h_name)} .= $h_val;
            }

            $h_name = $1;
            $h_val = $2;
        } elsif ($lines[$i] =~ /^\s+(.*)/) {
            $h_val =~ s/\s+$//;
            $h_val .= " $1";
        }
    }

    $txt = join("\n", $lines[$i .. scalar(@lines)]);

    return ($hdr, $txt);
}

