/*
 * xdcp.c:
 * Copy or move data in an xdata store.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 *
 */

static const char rcsid[] = "$Id: xdcp.c,v 1.3 2001/03/25 23:31:32 chris Exp $";
static const char copyright[] = "$Copyright: (c) 2001 Chris Lightfoot. All rights reserved.$";

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/file.h>

#include "xdata.h"

const char optstring[] = "+hp:t:f";

int move = 0;

void usage() {
    fprintf(stderr, move ? 
"xdmv: rename a key in an xdata database\n" :
"xdcp: copy a key in an xdata database\n");
    fprintf(stderr,
"\n"
"usage: %s [-h] [-p path] [-t type] [-f] old-key new-key\n"
"\n"
"   -h        Displays a help message.\n"
"\n"
"   -p path   Specified the path in which to look for xdata databases; if\n"
"             not specified, path is taken from the environment variable\n"
"             XDPATH\n"
"\n"
"   -t type   Specifies the type of data; if not specified, type is taken\n"
"             from the environment variable XDTYPE.\n"
"\n"
"   -f        Overwrite an existing key, if present.\n"
"\n"
"The datum associated with old-key is %s to new-key.\n"
"\n"
"If the datum is %s, %s will exit successfully.\n"
"\n"
"Copyright (c) 2001 Chris Lightfoot. Distribute freely.\n",
        move ? "xdmv" : "xdcp", move ? "moved" : "copied",
        move ? "moved" : "copied", move ? "xdmv" : "xdcp");

    exit(1);
}

int main(int argc, char *argv[]) {
    const char *type = getenv("XDTYPE"), *path = getenv("XDPATH"), *oldkey, *newkey;
    int overwrite = 0;
    int c;
    char *data;
    size_t len;
    XDATA xd;
    int ret = 1;
    char *name = "xdcp";

    if (strstr(argv[0], "xdmv")) {
        move = 1;
        name = "xdmv";
    }

    opterr = 0;
    while ((c = getopt(argc, argv, optstring)) != EOF) {
        switch (c) {
            case 'h':
                usage();
                break;
                
            case 'p':
                if (!optarg) {
                    fprintf(stderr, "%s: option -p requires an argument\n", name);
                    usage();
                } else path = optarg;
                break;

            case 't':
                if (!optarg) {
                    fprintf(stderr, "%s: option -t requires an argument\n", name);
                    usage();
                } else type = optarg;
                break;

            case 'f':
                overwrite = 1;
                break;
                
            default:
                fprintf(stderr, "%s: unrecognised option -%c\n", name, optopt);
                usage();
        }
    }

    if (!path) {
        fprintf(stderr, "%s: no xdata path specified\n", name);
        usage();
    }

    if (!type) {
        fprintf(stderr, "%s: no xdata type specified\n", name);
        usage();
    }
    
    oldkey = argv[optind++];
    newkey = argv[optind++];
    if (!oldkey || !newkey) {
        fprintf(stderr, "%s: not enough keys specified\n", name);
        usage();
    }

    /* OK, try to get hold of data. */
    xd = xdata_open(path);
    if (!xd) {
        fprintf(stderr, "%s: cannot open xdata\n", name);
        exit(1);
    }

    if (!xdata_set_type(xd, type, 1, NULL)) {
        fprintf(stderr, "%s: %s\n", name, xd->errmsg);
        xdata_close(xd);
        exit(1);
    }
    
    data = xdata_get(xd, oldkey, &len, NULL);
    if (!data) {
        if (xd->status)
            fprintf(stderr, "%s: key `%s': %s\n", name, oldkey, xd->errmsg);
        else
            fprintf(stderr, "%s: key `%s': no such key\n", name, oldkey);
        xdata_close(xd);
        exit(1);
    }

    if (!overwrite && xdata_exists(xd, newkey)) {
        fprintf(stderr, "%s: key `%s': already exists\n", name, newkey);
        xdata_close(xd);
        exit(1);
    }



    /* Now insert it into the database. */
    if (!xdata_put(xd, newkey, data, len, time(NULL), overwrite))
        fprintf(stderr, "%s: key `%s': %s\n", name, newkey, xd->errmsg);
    else ret = 0;


    if (move && !xdata_remove(xd, oldkey, time(NULL)))
        fprintf(stderr, "%s: key `%s': %s\n", name, oldkey, xd->errmsg);
    
    free(data);

    xdata_close(xd);
    exit(ret);
}
