/*
 * xded.c:
 * Edit data in an xdata database.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 *
 */

static const char rcsid[] = "$Id: xded.c,v 1.4 2001/09/15 15:04:45 chris Exp $";
static const char copyright[] = "$Copyright: (c) 2001 Chris Lightfoot. All rights reserved.$";

#include <sys/types.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "util.h"
#include "xdata.h"

#ifndef XDATA_EDITOR
#   define XDATA_EDITOR "/bin/vi"
#endif

const char optstring[] = "+hp:t:n";

void usage() {
    fprintf(stderr,
"xded: edit a datum in an xdata database\n"
"\n"
"usage: xded [-h] [-p path] [-t type] key\n"
"\n"
"   -h        Displays a help message.\n"
"\n"
"   -p path   Specified the path in which to look for xdata databases; if\n"
"             not specified, path is taken from the environment variable\n"
"             XDPATH\n"
"\n"
"   -t type   Specifies the type of data; if not specified, type is taken\n"
"             from the environment variable XDTYPE.\n"
"\n"
"   -n        Create a new blank key if the named key does not exist.\n"
"\n"
"The datum associated with the given key will be written out to a temporary\n"
"file, and an editor invoked on the file. The value of the environment\n"
"variable EDITOR can be used to specify which editor is used; by default,\n"
XDATA_EDITOR" is used.\n"
"\n"
"If the data are edited, xded will exit successfully.\n"
"\n"
"Copyright (c) 2001 Chris Lightfoot. Distribute freely.\n");

    exit(1);
}

int main(int argc, char *argv[]) {
    const char *type = getenv("XDTYPE"), *path = getenv("XDPATH"), *key;
    const char *editor = getenv("EDITOR");
    int c;
    char *data, *p;
    size_t len, buf_len, l;
    XDATA xd;
    int ret = 1;
    int fd;
    char tempfile[] = "xded-temp-XXXXXX";
    int editok = 0, makenew = 0;

    if (!editor) editor = XDATA_EDITOR;
    
    opterr = 0;
    while ((c = getopt(argc, argv, optstring)) != EOF) {
        switch (c) {
            case 'h':
                usage();
                break;
                
            case 'p':
                if (!optarg) {
                    fprintf(stderr, "xded: option -p requires an argument\n");
                    usage();
                } else path = optarg;
                break;

            case 't':
                if (!optarg) {
                    fprintf(stderr, "xded: option -t requires an argument\n");
                    usage();
                } else type = optarg;
                break;

            case 'n':
                makenew = 1;
                break;

            default:
                fprintf(stderr, "xded: unrecognised option -%c\n", optopt);
                usage();
        }
    }

    if (!path) {
        fprintf(stderr, "xded: no xdata path specified\n");
        usage();
    }

    if (!type) {
        fprintf(stderr, "xded: no xdata type specified\n");
        usage();
    }
    
    key = argv[optind++];
    if (!key) {
        fprintf(stderr, "xded: no key specified\n");
        usage();
    }

    /* OK, try to get hold of data. */
    xd = xdata_open(path);
    if (!xd) {
        fprintf(stderr, "xded: cannot open xdata\n");
        exit(1);
    }

    if (!xdata_set_type(xd, type, 1, NULL)) {
        fprintf(stderr, "xded: %s\n", xd->errmsg);
        xdata_close(xd);
        exit(1);
    }
    
    data = xdata_get(xd, key, &len, NULL);
    if (!data && (xd->status || !makenew)) {
        if (xd->status)
            fprintf(stderr, "xded: key `%s': %s\n", key, xd->errmsg);
        else
            fprintf(stderr, "xded: key `%s': no such key\n", key);
        xdata_close(xd);
        exit(1);
    }

    /* Write it out to a temporary file, then invoke the editor on same. */
    fd = mkstemp(tempfile);
    if (fd == -1) {
        fprintf(stderr, "xded: mkstemp: %s\n", strerror(errno));
        xdata_close(xd);
        exit(1);
    } else {
        pid_t ed_pid;
        struct stat st1, st2;
        if (data) {
            if (xwrite(fd, data, len) == -1) {
                fprintf(stderr, "xded: write: %s\n", strerror(errno));
                goto tempfailed;
            }
        } else {
            if (xwrite(fd, "\n", 1) == -1) {
                fprintf(stderr, "xded: write: %s\n", strerror(errno));
                goto tempfailed;
            }
        }

        free(data);
        close(fd);
        
        if (stat(tempfile, &st1) == -1) {
            fprintf(stderr, "xded: stat: %s\n", strerror(errno));
            goto tempfailed;
        }
        
        ed_pid = fork();

        if (ed_pid == -1) {
            fprintf(stderr, "xded: fork: %s\n", strerror(errno));
            goto tempfailed;
        } else if (ed_pid == 0) {
            char *ed_argv[3];
            ed_argv[0] = (char*)editor;
            ed_argv[1] = tempfile;
            ed_argv[2] = NULL;

            execvp(editor, ed_argv);
            exit(1);    /* Exit on failure to exec. */
        } else {
            int status;
            
            if (wait(&status) == -1) {
                fprintf(stderr, "xded: wait: %s\n", strerror(errno));
                goto tempfailed;
            }
            
            if (stat(tempfile, &st2) == -1) {
                fprintf(stderr, "xded: stat: %s\n", strerror(errno));
                goto tempfailed;
            }

            if (st2.st_mtime == st1.st_mtime) {
                fprintf(stderr, "xded: file was not modified\n");
                goto tempfailed;
            }

            editok = 1;
        }

        fd = open(tempfile, O_RDONLY);
        if (fd == -1) {
            fprintf(stderr, "xded: %s: %s\n", tempfile, strerror(errno));
            goto tempfailed;
        }

        /* Read data from fd. */
        buf_len = 1024;
        len = 0;
        p = data = (char*)malloc(buf_len);
        do {
            l = xread(fd, p, buf_len - (data - p));
            if (l == -1 && errno != EINTR) {
                /* XXX what about EPIPE? */
                fprintf(stderr, "xded: %s: %s\n", tempfile, strerror(errno));
                xdata_close(xd);
                exit(1);
            }
            if (l > 0) {
                len += l;
                p += l;
                if (p == data + buf_len) {
                    l = p - data;
                    data = realloc(data, buf_len * 2);
                    buf_len *= 2;
                    p = data + l;
                }
            }
        } while (l);

        editok = 1;

tempfailed:
        unlink(tempfile);
        {
            /* There may be an editor temporary file; delete it. */
            char *bf = xsprintf("%s~", tempfile);
            unlink(bf);
            free(bf);
        }
        if (!editok) {
            xdata_close(xd);
            exit(1);
        }
    }

    /* Now insert it into the database. */
    if (!xdata_put(xd, key, data, len, time(NULL), 1))
        fprintf(stderr, "xded: key `%s': %s\n", key, xd->errmsg);
    else ret = 0;
    
    free(data);

    xdata_close(xd);
    exit(ret);
}
